
from .PSStackParser import LIT

##  Constants
##
LITERAL_PDF = LIT("PDF")
LITERAL_TEXT = LIT("Text")
LITERAL_FONT = LIT("Font")
LITERAL_FORM = LIT("Form")
LITERAL_IMAGE = LIT("Image")

COMMON_FONT_ATTRIBUTES = [
    "Type",
    "FontName",
    "FontFamily",
    "FontStretch",
    "FontWeight",
    "Flags", 
    "FontBBox", 
    "ItalicAngle", 
    "Ascent", 
    "Descent", 
    "Leading",
    "CapHeight", 
    "XHeight",
    "StemV", 
    "StemH", 
    "AvgWidth",
    "MaxWidth",
    "MissingWidth",
    # "FontFile",
    # "FontFile2",
    # "FontFile3",
    "Charset",
]


import zlib

from ..PSStackParser import PSException
from ..PSStackParser import PSObject
from ...utils import lzwdecode, settings

from .constants import * # pylint: disable=W0614
from .types import * # pylint: disable=W0614
from .exceptions import * # pylint: disable=W0614

##  PDFObjRef
##
class PDFObjRef(PDFObject):

    def __init__(self, doc, objid, a):
        if objid == 0:
            if settings.STRICT:
                raise PDFValueError('PDF object id cannot be 0.')
        self.doc = doc
        self.objid = objid
        #self.genno = genno  # Never used.
        return

    def __repr__(self):
        return '<PDFObjRef:%d>' % (self.objid)

    def resolve(self, default=None):
        try:
            return self.doc.getobj(self.objid)
        except PDFObjectNotFound:
            return default




import collections
from .PSStackParser import LIT

##  PDFColorSpace
##
LITERAL_DEVICE_GRAY = LIT('DeviceGray')
LITERAL_DEVICE_RGB = LIT('DeviceRGB')
LITERAL_DEVICE_CMYK = LIT('DeviceCMYK')


class PDFColorSpace(object):

    def __init__(self, name: str, ncomponents: int):
        self.name: str = name
        self.ncomponents: int = ncomponents
        return

    def __repr__(self):
        return '<PDFColorSpace: %s, ncomponents=%d>' % (self.name, self.ncomponents)


PREDEFINED_COLORSPACE = collections.OrderedDict()

for (name, n) in [
    ('DeviceGray', 1),  # default value first
    ('CalRGB', 3),
    ('CalGray', 1),
    ('Lab', 3),
    ('DeviceRGB', 3),
    ('DeviceCMYK', 4),
    ('Separation', 1),
    ('Indexed', 1),
    ('Pattern', 1),
]:
    PREDEFINED_COLORSPACE[name]=PDFColorSpace(name, n)



import re
from .types import PSLiteral, PSSymbolTable, PSKeyword

PSLiteralTable = PSSymbolTable(PSLiteral)
PSKeywordTable = PSSymbolTable(PSKeyword)
LIT = PSLiteralTable.intern
KWD = PSKeywordTable.intern
KEYWORD_PROC_BEGIN = KWD(b'{')
KEYWORD_PROC_END = KWD(b'}')
KEYWORD_ARRAY_BEGIN = KWD(b'[')
KEYWORD_ARRAY_END = KWD(b']')
KEYWORD_DICT_BEGIN = KWD(b'<<')
KEYWORD_DICT_END = KWD(b'>>')

##  PSBaseParser
##
EOL = re.compile(br'[\r\n]')
SPC = re.compile(br'\s')
NONSPC = re.compile(br'\S')
HEX = re.compile(br'[0-9a-fA-F]')
END_LITERAL = re.compile(br'[#/%\[\]()<>{}\s]')
END_HEX_STRING = re.compile(br'[^\s0-9a-fA-F]')
HEX_PAIR = re.compile(br'[0-9a-fA-F]{2}|.')
END_NUMBER = re.compile(br'[^0-9]')
END_KEYWORD = re.compile(br'[#/%\[\]()<>{}\s]')
END_STRING = re.compile(br'[()\134]')
OCT_STRING = re.compile(br'[0-7]')
ESC_STRING = {b'b': 8, b't': 9, b'n': 10, b'f': 12, b'r': 13, b'(': 40, b')': 41, b'\\': 92}



class PDFGraphicStateColor:
    class NotSet(Exception): pass

    def __init__(self):
        self.__rgb = None
        self.__gray = None
        self.__cmyk = None
        self.__custom = None

    def get_color(self):
        if self.__rgb is not None:
            return 'rgb', self.rgb
        elif self.__cmyk is not None:
            return 'cmyk', self.cmyk
        elif self.__gray is not None:
            return 'gray', self.gray
        elif self.__custom is not None:
            return 'custom', self.custom
        else: return None, None
    
    @property
    def is_set(self):
        return self.__rgb is not None or self.__cmyk is not None or self.__gray is not None

    @property
    def rgb(self):
        if self.__rgb is not None: return [*self.__rgb]
        else: raise self.NotSet

    @property
    def cmyk(self):
        if self.__cmyk is not None: return [*self.__cmyk]
        else: raise self.NotSet

    @property
    def gray(self):
        if self.__gray is not None: return float(self.__gray)
        else: raise self.NotSet

    @property
    def custom(self):
        if self.__custom is not None: return dict(self.__custom)
        else: raise self.NotSet

    @rgb.setter
    def rgb(self, rgb):
        self.__rgb = rgb

    @cmyk.setter
    def cmyk(self, cmyk):
        self.__cmyk = cmyk

    @gray.setter
    def gray(self, color):
        self.__gray = color

    @custom.setter
    def custom(self, color):
        self.__custom = color

    def copy(self):
        cp = self.__class__()
        cp.gray = self.__gray
        cp.rgb = self.__rgb
        cp.cmyk = self.__cmyk
        cp.custom = self.__custom
        return cp

    def __repr__(self):
        rep = "<PDFGraphicStateColor"
        if self.__gray is not None: rep += f' gray={self.__gray}'
        if self.__rgb is not None: rep += f' rgb=[{",".join(map(str, self.__rgb))}]'
        if self.__cmyk is not None: rep += f' cmyk=[{",".join(map(str, self.__cmyk))}]'
        if self.__custom is not None: rep += f' custom={self.__custom}'
        return rep + '/>'



# -*- coding: utf-8 -*-

from .PDFFont import PDFUnicodeNotDefined

from ..utils import isnumber, apply_matrix_pt
from ..utils import Bbox, mult_matrix, translate_matrix

from .PDFResourceManager import PDFResourceManager
from .PDFGraphicState import PDFGraphicState
from .PDFTextState import PDFTextState
from .PDFColorSpace import PDFColorSpace
from .PDFFont import PDFFont

##  PDFDevice
##
class PDFDevice(object):

    def __init__(self, rsrcmgr: PDFResourceManager):
        self.rsrcmgr = rsrcmgr
        self.ctm = None
        return

    def __repr__(self):
        return '<PDFDevice>'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        return

    def set_ctm(self, ctm):
        self.ctm = ctm
        return

    def begin_tag(self, tag, props=None):
        return

    def end_tag(self):
        return

    def do_tag(self, tag, props=None):
        return

    def begin_page(self, page, ctm):
        return

    def end_page(self, page):
        return

    def begin_figure(self, name, bbox: Bbox, matrix):
        return

    def end_figure(self, name):
        return

    def paint_path(self, graphicstate: PDFGraphicState, stroke, fill, evenodd, path):
        return

    def render_image(self, name, stream):
        return

    def render_string(self, textstate, seq, ncs, graphicstate: PDFGraphicState):
        return


##  PDFTextDevice
##
class PDFTextDevice(PDFDevice):

    def render_string(self, textstate: PDFTextState, seq: bytearray, ncs: PDFColorSpace, graphicstate: PDFGraphicState):
        textstate.matrix = mult_matrix(textstate.matrix, self.ctm)
        textstate.scaling = textstate.scaling * .01
        textstate.charspace = textstate.charspace * textstate.scaling
        textstate.wordspace = textstate.wordspace * textstate.scaling
       
        if textstate.font.is_multibyte():
            textstate.wordspace = 0
        dxscale = .001 * textstate.fontsize * textstate.scaling

        if textstate.font.is_vertical():
            self.__render_string_along(1,
                seq, textstate, dxscale, ncs, graphicstate
            )
        else:
            self.__render_string_along(0,
                seq, textstate, dxscale, ncs, graphicstate
            )

    def __render_string_along(self, idx: int, seq: bytearray, 
        textstate: PDFTextState, dxscale: float, ncs: PDFColorSpace, graphicstate: PDFGraphicState):
        needcharspace = False
        char_meta_datas = []
        for obj in seq:
            if isnumber(obj):
                textstate.linematrix[idx] -= obj*dxscale
                needcharspace = True
            else:
                for cid in textstate.font.decode(obj):
                    if needcharspace:
                        textstate.linematrix[idx] += textstate.charspace
                    
                    text = textstate.font.to_unichr(cid)
                    assert isinstance(text, str), str(type(text))

                    matrix = translate_matrix(textstate.matrix, textstate.linematrix)
                    adv, bbox = self.__compute_char_bbox(matrix, cid, textstate)
                    textstate.linematrix[idx] += adv

                    char_meta_datas.append(
                        ( text, bbox, )
                    )
                    if cid == 32 and textstate.wordspace:
                        textstate.linematrix[idx] += textstate.wordspace
                    needcharspace = True
                    
        self.render_char_box(char_meta_datas, textstate, graphicstate, ncs)

    def __compute_char_bbox(self, matrix, char_id: int, textstate: PDFTextState):
        font = textstate.font
        fontsize = textstate.fontsize
        rise = textstate.rise
        scaling = textstate.scaling

        textwidth = font.char_width(char_id)
        textdisp = font.char_disp(char_id)
        adv = textwidth * fontsize * scaling

        # compute the boundary rectangle.
        if font.is_vertical():
            # vertical
            width = font.get_width() * fontsize
            (vx, vy) = textdisp
            if vx is None:
                vx = width * 0.5
            else:
                vx = vx * fontsize * .001
            vy = (1000 - vy) * fontsize * .001
            tx = -vx
            ty = vy + rise
            bll = (tx, ty+adv)
            bur = (tx+width, ty)
        else:
            # horizontal
            height = font.get_height() * fontsize
            descent = font.get_descent() * fontsize
            ty = descent + rise
            bll = (0, ty)
            bur = (adv, ty+height)
        (x0, y0) = apply_matrix_pt(matrix, bll)
        (x1, y1) = apply_matrix_pt(matrix, bur)
        if x1 < x0:
            (x0, x1) = (x1, x0)
        if y1 < y0:
            (y0, y1) = (y1, y0)
        
        return (adv, ((x0, y0), (x1, y1)))

    def render_char_box(self, chars: list, textstate: PDFTextState, graphicstate: PDFGraphicState, ncs: PDFColorSpace):
        pass



import re
from logging import getLogger

from ...utils import settings, choplist, nunpack
from ..PSStackParser import PSEOF
from ..PSStackParser import KWD
from ..PDFParser import PDFStreamParser
from ..PDFStream import PDFStream
from ..PDFStream import dict_value

from .constants import LITERAL_OBJSTM, LITERAL_XREF
from .exceptions import PDFNoValidXRef, PDFSyntaxError


log = getLogger(__name__)

##  XRefs
##
class PDFBaseXRef(object):

    def get_trailer(self):
        raise NotImplementedError

    def get_objids(self):
        return []

    # Must return
    #     (strmid, index, genno)
    #  or (None, pos, genno)
    def get_pos(self, objid):
        raise KeyError(objid)


##  PDFXRef
##
class PDFXRef(PDFBaseXRef):

    def __init__(self):
        self.offsets = {}
        self.trailer = {}
        return

    def __repr__(self):
        return '<PDFXRef: offsets=%r>' % (self.offsets.keys())

    def load(self, parser):
        while True:
            try:
                (pos, line) = parser.nextline()
                if not line.strip():
                    continue
            except PSEOF:
                raise PDFNoValidXRef('Unexpected EOF - file corrupted?')
            if not line:
                raise PDFNoValidXRef('Premature eof: %r' % parser)
            if line.startswith(b'trailer'):
                parser.seek(pos)
                break
            f = line.strip().split(b' ')
            if len(f) != 2:
                raise PDFNoValidXRef('Trailer not found: %r: line=%r' % (parser, line))
            try:
                (start, nobjs) = map(int, f)
            except ValueError:
                raise PDFNoValidXRef('Invalid line: %r: line=%r' % (parser, line))
            for objid in range(start, start+nobjs):
                try:
                    (_, line) = parser.nextline()
                except PSEOF:
                    raise PDFNoValidXRef('Unexpected EOF - file corrupted?')
                f = line.strip().split(b' ')
                if len(f) != 3:
                    raise PDFNoValidXRef('Invalid XRef format: %r, line=%r' % (parser, line))
                (pos, genno, use) = f
                if use != b'n':
                    continue
                self.offsets[objid] = (None, int(pos), int(genno))
        log.info('xref objects: %r', self.offsets)
        self.load_trailer(parser)
        return

    def load_trailer(self, parser):
        try:
            (_, kwd) = parser.nexttoken()
            assert kwd is KWD(b'trailer'), str(kwd)
            (_, dic) = parser.nextobject()
        except PSEOF:
            x = parser.pop(1)
            if not x:
                raise PDFNoValidXRef('Unexpected EOF - file corrupted')
            (_, dic) = x[0]
        self.trailer.update(dict_value(dic))
        log.debug('trailer=%r', self.trailer)
        return

    def get_trailer(self):
        return self.trailer

    def get_objids(self):
        return iter(self.offsets.keys())

    def get_pos(self, objid):
        try:
            return self.offsets[objid]
        except KeyError:
            raise


##  PDFXRefFallback
##
class PDFXRefFallback(PDFXRef):

    def __repr__(self):
        return '<PDFXRefFallback: offsets=%r>' % (self.offsets.keys())

    PDFOBJ_CUE = re.compile(r'^(\d+)\s+(\d+)\s+obj\b')

    def load(self, parser):
        parser.seek(0)
        while 1:
            try:
                (pos, line) = parser.nextline()
            except PSEOF:
                break
            if line.startswith(b'trailer'):
                parser.seek(pos)
                self.load_trailer(parser)
                log.info('trailer: %r', self.trailer)
                break
            line=line.decode('latin-1') #default pdf encoding
            m = self.PDFOBJ_CUE.match(line)
            if not m:
                continue
            (objid, genno) = m.groups()
            objid = int(objid)
            genno = int(genno)
            self.offsets[objid] = (None, pos, genno)
            # expand ObjStm.
            parser.seek(pos)
            (_, obj) = parser.nextobject()
            if isinstance(obj, PDFStream) and obj.get('Type') is LITERAL_OBJSTM:
                stream = PDFStream.validated_stream(obj)
                try:
                    n = stream['N']
                except KeyError:
                    if settings.STRICT:
                        raise PDFSyntaxError('N is not defined: %r' % stream)
                    n = 0
                parser1 = PDFStreamParser(stream.get_data())
                objs = []
                try:
                    while 1:
                        (_, obj) = parser1.nextobject()
                        objs.append(obj)
                except PSEOF:
                    pass
                n = min(n, len(objs)//2)
                for index in range(n):
                    objid1 = objs[index*2]
                    self.offsets[objid1] = (objid, index, 0)
        return

##  PDFXRefStream
##
class PDFXRefStream(PDFBaseXRef):

    def __init__(self):
        self.data = None
        self.entlen = None
        self.fl1 = self.fl2 = self.fl3 = None
        self.ranges = []
        return

    def __repr__(self):
        return '<PDFXRefStream: ranges=%r>' % (self.ranges)

    def load(self, parser):
        (_, objid) = parser.nexttoken()  # ignored
        (_, genno) = parser.nexttoken()  # ignored
        (_, kwd) = parser.nexttoken()
        (_, stream) = parser.nextobject()
        if not isinstance(stream, PDFStream) or stream['Type'] is not LITERAL_XREF:
            raise PDFNoValidXRef('Invalid PDF stream spec.')
        size = stream['Size']
        index_array = stream.get('Index', (0, size))
        if len(index_array) % 2 != 0:
            raise PDFSyntaxError('Invalid index number')
        self.ranges.extend(choplist(2, index_array))
        (self.fl1, self.fl2, self.fl3) = stream['W']
        self.data = stream.get_data()
        self.entlen = self.fl1+self.fl2+self.fl3
        self.trailer = stream.attrs
        log.info('xref stream: objid=%s, fields=%d,%d,%d',
                 ', '.join(map(repr, self.ranges)),
                 self.fl1, self.fl2, self.fl3)
        return

    def get_trailer(self):
        return self.trailer

    def get_objids(self):
        for (start, nobjs) in self.ranges:
            for i in range(nobjs):
                offset = self.entlen * i
                ent = self.data[offset:offset+self.entlen]
                f1 = nunpack(ent[:self.fl1], 1)
                if f1 == 1 or f1 == 2:
                    yield start+i
        return

    def get_pos(self, objid):
        index = 0
        for (start, nobjs) in self.ranges:
            if start <= objid and objid < start+nobjs:
                index += objid - start
                break
            else:
                index += nobjs
        else:
            raise KeyError(objid)
        offset = self.entlen * index
        ent = self.data[offset:offset+self.entlen]
        f1 = nunpack(ent[:self.fl1], 1)
        f2 = nunpack(ent[self.fl1:self.fl1+self.fl2])
        f3 = nunpack(ent[self.fl1+self.fl2:])
        if f1 == 1:
            return (None, f2, f3)
        elif f1 == 2:
            return (f2, f3, 0)
        else:
            # this is a free object
            raise KeyError(objid)



import logging

from typing import Dict

from contextlib import contextmanager

from ..layouts import LTPage
from ..layouts import LTItem
from ..layouts import LTCurve, LTLine, LTRect, CurvePath
from ..layouts import LTFigure
from ..layouts import LTImage
from ..layouts import LTChar, LTCharBlock
from ..utils import enc

from .PDFConverter import PDFConverter
from ..interpreter.PDFGraphicState import PDFGraphicStateColor

log = logging.getLogger(__name__)

def get_color(col: PDFGraphicStateColor):
    color_type, color_val = col.get_color()
    if color_type is not None:
        if color_type == 'rgb':
            return f"rgb({','.join(map(str, color_val))})"
        elif color_type == 'cmyk':
            return f"cmyk({','.join(map(str, color_val))})"
        elif color_type == 'gray':
            return f'rgba(0,0,0,{1-color_val})'
        elif color_type == "custom":
            # Todo: convert these to more friendly color types
            return f"custom({col.custom['type']})"
    return "none"

class HTMLConverter(PDFConverter):
    def __init__(self, 
        rsrcmgr, 
        outfp, 
        codec='utf-8', 
        pageno=1, 
        imagewriter=None
    ):
        PDFConverter.__init__(self, rsrcmgr, outfp, imagewriter=imagewriter, codec=codec, pageno=pageno)
        self.__levels_deep = 0
        self.__current_height = None
        self.__current_width = None
        self.write_header()
    
    def write(self, text: str, lineend: str = '\n', deep_space: str = ' '):
        text = deep_space*self.__levels_deep + text + lineend
        self.write_raw(text)
    
    def write_raw(self, text):
        if self.codec:
            text = text.encode(self.codec)
        self.outfp.write(text)
    
    def write_header(self):
        self.write('<html>')
        self.__levels_deep += 1
        self.write('<head>')
        self.__levels_deep += 1
        meta_attr = {
            'http-equiv': "Content-Type",
            'content': "text/html",
        }
        if self.codec:
            meta_attr['charset'] = self.codec

        self.place_elm_close('meta', attr_dict=meta_attr, as_childless=False)
        with self.place_elm_with_child('style'):
            self.write('''.page { position: relative; overflow: hidden;border: 1px black solid;}''')
        self.__levels_deep -= 1
        self.write('</head>')
        self.write('<body>')
        self.__levels_deep += 1
    
    def write_footer(self):
        self.write('</body>')
        self.__levels_deep -= 1
        self.write('</html>')
        self.__levels_deep -= 1

    @contextmanager
    def place_elm_with_child(self, tag_name: str, attr_dict: Dict[str, str] = {}, css: Dict[str, str] = None, no_additional_char = False):
        lineend = '\n'
        deep_space = ' '
        if no_additional_char:
            lineend = ''
            deep_space = ''

        if css is not None:
            attr_dict['style'] = ";".join([
                f'{name}: {value}'
                for name, value in css.items()
            ])
        attrs = " ".join([
            f'{name}="{value}"'
            for name, value in attr_dict.items()
        ])
        self.write(f"<{tag_name} {attrs}>", lineend=lineend)
        self.__levels_deep += 1
        yield
        self.__levels_deep -= 1
        self.write(f"</{tag_name}>", deep_space=deep_space)

    def place_elm_close(self, tag_name: str, attr_dict: Dict[str, str] = {}, css: Dict[str, str] = None, as_childless = True):
        if css is not None:
            attr_dict['style'] = ";".join([
                f'{name}: {value}'
                for name, value in css.items()
            ])
        attrs = " ".join([
            f'{name}="{value}"'
            for name, value in attr_dict.items()
        ])
        if as_childless:
            self.write(f"<{tag_name} {attrs}></{tag_name}>")
        else:
            self.write(f"<{tag_name} {attrs}/>")
    
    def place_image(self, item: LTImage):
        name = item.name
        if self.imagewriter is not None:
            name = self.imagewriter.export_image(item)
            self.place_elm_close('img', {
                'src': enc(name, None),
                'width': f"{item.width}px", 
                'height': f"{item.height}px",
            })
        return

    def render_curve(self, item: LTCurve):
        css = {
            'position': 'absolute', 
            'left': f'{item.x0}px', 
            'bottom': f'{item.y0}px', 
        }
        path = ""
        svg_attr = {
            'class': 'curve',
            'z-index': 2
        }

        if item.height > 0:
            dy = 1.0/item.height
            dx = (item.width/item.height)/(item.x1-item.x0)
            svg_attr.update({
                "viewBox": "0 0 %s %s" % ( item.width/item.height, 1 )
            })
        else:
            dx = 1.0/(item.x1-item.x0)
            dy = 1.0
            svg_attr.update({
                "viewBox": "0 0 1 1"
            })

        if item.height > 1:
            svg_attr.update({
                'height': f'{item.height}px',
            })
        else:
            svg_attr.update({
                'height': '1px',
            })
        
        if item.width > 1:
            svg_attr.update({
                'width': f'{item.width}px',
            })
        else:
            svg_attr.update({
                'width': '1px',
            })

        for p in item.paths:
            if p.method == CurvePath.METHOD.MOVE_TO:
                path += f'M{" ".join( f"{(item.width - (p.x-item.x0) )*dx} {(item.height- (p.y-item.y0) )*dy}" for p in p.points )} '
            if p.method == CurvePath.METHOD.LINE_TO:
                path += f'L{" ".join( f"{(item.width - (p.x-item.x0) )*dx} {( item.height - (p.y-item.y0) )*dy}" for p in p.points )} '
            if p.method == CurvePath.METHOD.CLOSE_PATH:
                path += 'Z'
        with self.place_elm_with_child('svg', svg_attr, css):
            self.place_elm_close('path', {
                "d": path,
                'stroke': get_color(item.stroke),
                'fill': get_color(item.fill),
            })

    def receive_layout(self, ltpage: LTPage):
        def render(item: LTItem):
            if isinstance(item, LTPage):
                with self.place_elm_with_child('div', { "class": 'page', "id": f"page-{item.pageid}" }, { 'width': f'{item.width}px', 'height': f'{item.height}px' }):
                    self.__current_height = item.height
                    self.__current_width = item.width
                    for child in item:
                        render(child)
            elif isinstance(item, LTCurve):
                self.render_curve(item)
            elif isinstance(item, LTFigure):
                with self.place_elm_with_child('div', { "class": 'figure' }, {
                    'position': 'absolute', 
                    'left': f'{item.x0}px', 
                    'bottom': f'{item.y0}px', 
                    'width': f'{item.width}px', 
                    'height': f'{item.height}px',
                }):
                    for child in item:
                        render(child)
            elif isinstance(item, LTImage):
                self.place_image(item)
            elif isinstance(item, LTChar):
                with self.place_elm_with_child('span', { 'class': 'char' }, {
                    'width': item.width,
                    'min-width': item.width,
                    'max-width': item.width,
                    'height': item.height,
                    'min-height': item.height,
                    'max-height': item.height,
                }):
                    self.write(item.get_text())
            elif isinstance(item, LTCharBlock):
                with self.place_elm_with_child('span', { 'class': 'char-block' }, {
                        'position': 'absolute',
                        'left': f'{item.x0}',
                        'bottom': f'{item.y0}',
                        'text-align': 'justify',
                        'font-size': f'{item.size}px',
                        'font-family': item.fontname,
                        'font-weight': item.font.font_weight,
                        'color': get_color(item.graphicstate.ncolor),
                        'width': item.width,
                        'min-width': item.width,
                        'max-width': item.width,
                        'height': item.height,
                        'min-height': item.height,
                        'max-height': item.height,
                        'transform': f'skew({item.font.italic_angle}deg, 0deg)',
                        'text-align': 'center',
                        'display': 'flex',
                        'flex-flow': 'row nowrap',
                        'justify-content': 'space-between',
                        # 'padding-left': item.font.leading
                    }):
                    for char in item:
                        render(char)
        render(ltpage)

    def close(self):
        self.write_footer()
    


import logging
import re
import io

from abc import abstractmethod
from typing import Tuple, List

from ._base import LTItem
from .LTPage import LTPage
from .LTCurves import LTCurve, LTLine, LTRect, LTHorizontalLine, LTVerticalLine
from .LTFigure import LTFigure
from .LTImage import LTImage
from .LTChar import LTChar
from .LTCharBlock import LTCharBlock

from ..utils import Point, Bbox, INF
from ..utils import apply_matrix_pt, mult_matrix

from ..interpreter.PDFFont import PDFUnicodeNotDefined
from ..interpreter.PDFResourceManager import PDFResourceManager
from ..interpreter.PDFGraphicState import PDFGraphicState, PDFGraphicStateColor
from ..interpreter.PDFdevice import PDFTextDevice
from ..interpreter.PDFTextState import PDFTextState
from ..interpreter.PDFColorSpace import PDFColorSpace
from ..interpreter.PDFPage import PDFPage
from ..interpreter.PDFStream import PDFStream
from ..interpreter.types import CurvePath

log = logging.getLogger(__name__)


##  PDFLayoutAnalyzer
##
class PDFLayoutAnalyzer(PDFTextDevice):


    def __init__(self, rsrcmgr: PDFResourceManager, pageno: int=1):
        PDFTextDevice.__init__(self, rsrcmgr)
        self.pageno = pageno
        self._stack = []
        return

    def begin_page(self, page: PDFPage, ctm: Tuple[float]):
        (x0, y0, x1, y1) = page.mediabox
        (x0, y0) = apply_matrix_pt(ctm, (x0, y0))
        (x1, y1) = apply_matrix_pt(ctm, (x1, y1))
        self.cur_item = LTPage(
            self.pageno, 
            Bbox( 0, 0, abs(x0-x1), abs(y0-y1) )
        )
        return

    def end_page(self, page: PDFPage):
        assert not self._stack, str(len(self._stack))
        assert isinstance(self.cur_item, LTPage), str(type(self.cur_item))
        # self.cur_item.analyze()
        self.pageno += 1
        self.receive_layout(self.cur_item)
        return

    def begin_figure(self, name: str, bbox: List[Point], matrix: List[Point]):
        self._stack.append(self.cur_item)

        matrix = mult_matrix(matrix, self.ctm)
        (x, y, w, h) = bbox

        self.cur_item = LTFigure(name, [
            apply_matrix_pt(matrix, (p, q))
            for (p, q) in ((x, y), (x+w, y), (x, y+h), (x+w, y+h))
        ], matrix)

    def end_figure(self, name: str):
        fig = self.cur_item
        assert isinstance(self.cur_item, LTFigure), str(type(self.cur_item))
        self.cur_item = self._stack.pop()
        self.cur_item.add(fig)
        return

    def render_image(self, name: str, stream: PDFStream):
        assert isinstance(self.cur_item, LTFigure), str(type(self.cur_item))
        item = LTImage(
            name, 
            stream,
            Bbox(
                self.cur_item.x0, self.cur_item.y0,
                self.cur_item.x1, self.cur_item.y1
            )
        )
        self.cur_item.add(item)
        return

    def paint_path(self, gstate: PDFGraphicState, evenodd: bool, paths: List[CurvePath]):
        shape = ''.join(x.method.value for x in paths)
        if shape == 'ml':
            # horizontal/vertical line
            pt0 = paths[0].points[0]
            pt1 = paths[1].points[0]
            (x0, y0) = apply_matrix_pt(self.ctm, (pt0.x, pt0.y))
            (x1, y1) = apply_matrix_pt(self.ctm, (pt1.x, pt1.y))
            if x0 == x1:
                self.cur_item.add(LTVerticalLine(
                    linewidth=gstate.linewidth, 
                    paths=paths,
                    bbox=Bbox( x0, y0, x1, y1 ),
                    stroke=gstate.scolor, 
                    fill=gstate.ncolor, 
                    evenodd=evenodd, 
                ))
            elif y0 == y1:
                self.cur_item.add(LTHorizontalLine(
                    linewidth=gstate.linewidth, 
                    paths=paths,
                    bbox=Bbox( x0, y0, x1, y1 ),
                    stroke=gstate.scolor, 
                    fill=gstate.ncolor, 
                    evenodd=evenodd, 
                ))
            else:
                self.cur_item.add(LTLine(
                    linewidth=gstate.linewidth, 
                    paths=paths,
                    bbox=Bbox( x0, y0, x1, y1 ),
                    stroke=gstate.scolor, 
                    fill=gstate.ncolor, 
                    evenodd=evenodd, 
                ))
        elif shape == 'mlllh':
            # rectangle
            pt0 = paths[0].points[0]
            pt1 = paths[1].points[0]
            pt2 = paths[2].points[0]
            pt3 = paths[3].points[0]
            (x0, y0) = apply_matrix_pt(self.ctm, (pt0.x, pt0.y))
            (x1, y1) = apply_matrix_pt(self.ctm, (pt1.x, pt1.y))
            (x2, y2) = apply_matrix_pt(self.ctm, (pt2.x, pt2.y))
            (x3, y3) = apply_matrix_pt(self.ctm, (pt3.x, pt3.y))
            if ((x0 == x1 and y1 == y2 and x2 == x3 and y3 == y0) or
                (y0 == y1 and x1 == x2 and y2 == y3 and x3 == x0)):
                self.cur_item.add(LTRect( 
                    linewidth=gstate.linewidth, 
                    paths=paths,
                    bbox=Bbox(
                        x0=min(x0,x1,x2,x3),
                        x1=max(x0,x1,x2,x3),
                        y0=min(y0,y1,y2,y3),
                        y1=max(y0,y1,y2,y3),
                    ),
                    evenodd=evenodd, 
                    stroke=gstate.scolor, 
                    fill=gstate.ncolor,
                ))
        else:
            (x0, y0, x1, y1) = (INF, INF, -INF, -INF)
            for path in paths:
                if path.method in [CurvePath.METHOD.MOVE_TO, CurvePath.METHOD.LINE_TO]:
                    x0 = min(x0, path.points[0].x)
                    y0 = min(y0, path.points[0].y)
                    x1 = max(x1, path.points[0].x)
                    y1 = max(y1, path.points[0].y)
                elif path.method in [CurvePath.METHOD.CURVE_BOTH_TO]:
                    x0 = min(x0, path.points[2].x)
                    y0 = min(y0, path.points[2].y)
                    x1 = max(x1, path.points[2].x)
                    y1 = max(y1, path.points[2].y)
                elif path.method in [CurvePath.METHOD.CURVE_NEXT_TO, CurvePath.METHOD.CURVE_FIRST_TO]:
                    x0 = min(x0, path.points[1].x)
                    y0 = min(y0, path.points[1].y)
                    x1 = max(x1, path.points[1].x)
                    y1 = max(y1, path.points[1].y)
            self.cur_item.add(LTCurve(
                linewidth=gstate.linewidth, 
                paths=paths,
                bbox=Bbox(x0,y0,x1,y1), 
                evenodd=evenodd,
                stroke=gstate.scolor, 
                fill=gstate.ncolor
            ))

    def render_char_box(self, chars: list, textstate: PDFTextState, graphicstate: PDFGraphicState, ncs: PDFColorSpace):
        self.cur_item.add(LTCharBlock(
            chars=chars, 
            textstate=textstate, 
            ncs=ncs, 
            graphicstate=graphicstate
        ))

    @abstractmethod
    def receive_layout(self, ltpage: LTPage):
        raise NotImplementedError



from typing import List

from ._base import LTContainer
from ..utils import Bbox, Point, matrix2str

##  LTFigure
##
class LTFigure(LTContainer):

    def __init__(self, name: str, pts: List[Point], matrix):
        LTContainer.__init__(self, Bbox.from_points(pts))
        self.name = name
        self.matrix = matrix

    def __repr__(self):
        return ('<%s(%s) %s matrix=%s>' %
                (self.__class__.__name__, self.name,
                 str(self.bbox), matrix2str(self.matrix)))
